pub mod config;
pub mod extractor;
pub mod model;
pub mod pages;
pub mod state;
pub mod templates;
pub mod token;

use std::net::{IpAddr, Ipv4Addr, SocketAddr};
use std::sync::Arc;

use axum::{routing::get, Router};
use clap::Parser;
use eyre::{Result, WrapErr};
use jsonwebtoken::{DecodingKey, EncodingKey};
use sqlx::migrate;
use sqlx::postgres::PgPoolOptions;
use tower_http::trace::TraceLayer;
use tracing::{debug, info};

use crate::config::Config;
use crate::state::AppState;
use crate::templates::handlebars;

#[tokio::main]
async fn main() -> Result<()> {
    dotenvy::dotenv().ok();

    tracing_subscriber::fmt()
        .with_max_level(tracing::Level::DEBUG)
        .init();

    debug!("Reading configuration");

    let config = Config::parse();

    let hbs = handlebars(config.dev_mode)?;

    debug!("Initializing database");

    let pool = PgPoolOptions::new()
        .max_connections(500)
        .connect(&config.database_url)
        .await
        .wrap_err("creating database pool")?;

    migrate!().run(&pool).await.wrap_err("running migrations")?;

    let state = Arc::new(AppState {
        pool,
        hbs,
        encoding_key: EncodingKey::from_base64_secret(&config.secret_key)
            .wrap_err("parsing $SECRET_KEY")?,
        decoding_key: DecodingKey::from_base64_secret(&config.secret_key)
            .wrap_err("parsing $SECRET_KEY")?,
    });

    debug!("Building router");

    let app = Router::new()
        .route("/", get(pages::index))
        .route("/users/:id", get(pages::get_user))
        .route(
            "/users/new",
            get(pages::get_user_new).post(pages::post_user_new),
        )
        .layer(TraceLayer::new_for_http())
        .with_state(state);

    info!("Listening on port {}", config.port);

    let socket = SocketAddr::new(IpAddr::V4(Ipv4Addr::new(0, 0, 0, 0)), config.port);
    axum::Server::bind(&socket)
        .serve(app.into_make_service())
        .await
        .wrap_err("running server")
}
