use std::{collections::BTreeMap, sync::Arc};

use axum::{
    body::BoxBody,
    extract::{Path, State},
    response::{IntoResponse, Response},
    Form,
};
use serde::Deserialize;
use sqlx::{query, query_as};
use tracing::{error, info};

use crate::{
    extractor::ExtractUserToken,
    model::user::{Role, User},
    state::AppState,
    templates::{Page, PageData},
};

pub async fn get_user(
    State(state): State<Arc<AppState>>,
    Path(user_id): Path<i32>,
) -> impl IntoResponse {
    let user = match query_as!(User, r#"SELECT * FROM users WHERE id = $1"#, user_id)
        .fetch_one(&state.pool)
        .await
    {
        Ok(user) => user,
        Err(why) => {
            error!("Getting user: {}", why);
            return Page::new(state.hbs.render(
                "error.hbs",
                &PageData {
                    message: Some("Internal server error".into()),
                    ..Default::default()
                },
            ));
        }
    };

    Page::new(state.hbs.render(
        "error.hbs",
        &PageData {
            user: Some(user),
            ..Default::default()
        },
    ))
}

pub async fn get_user_new(
    State(state): State<Arc<AppState>>,
    ExtractUserToken(user): ExtractUserToken,
) -> impl IntoResponse {
    // If there's already an authenticated user, check if they're an admin
    // Admins can create new users, others can't
    if let Some(user) = user {
        tracing::debug!("User is authenticated");

        if user.role == Role::Admin {
            tracing::debug!("Authenticated user is admin");

            return Page::new(
                state
                    .hbs
                    .render::<PageData>("new_user.hbs", &Default::default()),
            );
        }

        tracing::debug!("Authenticated user is not an admin");
        return Page::new(
            state
                .hbs
                .render::<PageData>("error.hbs", &Default::default()),
        );
    }

    tracing::debug!("There is no authenticated user");

    // Else, let the user create a new account, but only if no accounts exist yet
    let count = match query!(r#"SELECT COUNT(id) FROM users"#)
        .fetch_one(&state.pool)
        .await
    {
        Ok(r) => r.count.unwrap_or(0),
        Err(why) => {
            error!("Getting user count: {}", why);
            return Page::new(
                state
                    .hbs
                    .render::<PageData>("error.hbs", &Default::default()),
            );
        }
    };

    tracing::debug!("Current user count: {}", count);

    let mut map = BTreeMap::new();
    map.insert("count", count);

    Page::new(state.hbs.render("new_user.hbs", &map))
}

#[derive(Deserialize)]
pub struct UserNew {
    pub username: String,
    pub password: String,
    pub password2: String,
}

pub async fn post_user_new(
    State(state): State<Arc<AppState>>,
    ExtractUserToken(existing_user): ExtractUserToken,
    Form(form_data): Form<UserNew>,
) -> Response<BoxBody> {
    if existing_user.is_none() {
        let count = match query!(r#"SELECT COUNT(id) FROM users"#)
            .fetch_one(&state.pool)
            .await
        {
            Ok(r) => r.count.unwrap_or(0),
            Err(why) => {
                error!("Getting user count: {}", why);

                return Page::new(state.hbs.render(
                    "error.hbs",
                    &PageData {
                        message: Some("Internal server error".into()),
                        ..Default::default()
                    },
                ))
                .into_response();
            }
        };

        if count != 0 {
            info!("");

            return Page::new(state.hbs.render(
                "error.hbs",
                &PageData {
                    message: Some("You need to be logged in to create new users".into()),
                    ..Default::default()
                },
            ))
            .into_response();
        }
    } else if let Some(user) = existing_user.clone() {
        if user.role != Role::Admin {
            info!(
                user_id = user.id,
                "User tried to access POST /users/new despite not being admin"
            );

            return Page::new(state.hbs.render(
                "error.hbs",
                &PageData {
                    message: Some("Only admins can create new users".into()),
                    ..Default::default()
                },
            ))
            .into_response();
        }
    }

    if form_data.password != form_data.password2 {
        return Page::new(state.hbs.render(
            "new_user.hbs",
            &PageData {
                message: Some("Passwords don't match".into()),
                ..Default::default()
            },
        ))
        .into_response();
    }

    let hashed_password = match bcrypt::hash(form_data.password, 12) {
        Ok(hash) => hash,
        Err(why) => {
            error!("Hashing password: {}", why);

            return Page::new(state.hbs.render(
                "error.hbs",
                &PageData {
                    message: Some("Internal server error".into()),
                    ..Default::default()
                },
            ))
            .into_response();
        }
    };

    let user = match query_as!(
        User,
        r#"INSERT INTO users (username, password) VALUES ($1, $2) RETURNING *"#,
        form_data.username,
        hashed_password,
    )
    .fetch_one(&state.pool)
    .await
    {
        Ok(u) => u,
        Err(why) => {
            error!("Creating user: {}", why);

            return Page::new(state.hbs.render(
                "error.hbs",
                &PageData {
                    message: Some("Internal server error".into()),
                    ..Default::default()
                },
            ))
            .into_response();
        }
    };

    if existing_user.is_some() {
        // Render template with new user info (ID, username)

        return Page::new(state.hbs.render(
            "new_user_confirm.hbs",
            &PageData {
                user: existing_user,
                ..Default::default()
            },
        ))
        .into_response();
    }

    Page::new(state.hbs.render(
        "new_user.hbs",
        &PageData {
            ..Default::default()
        },
    ))
    .into_response()
}
